"""Ariadne 内置的 Dispatcher"""
from typing import TYPE_CHECKING, ContextManager, TypedDict

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.entities.event import Dispatchable
from graia.broadcast.interfaces.dispatcher import DispatcherInterface

from .context import (
    adapter_ctx,
    ariadne_ctx,
    broadcast_ctx,
    enter_context,
    event_loop_ctx,
)
from .message.chain import MessageChain
from .message.element import Source

if TYPE_CHECKING:
    from .app import Ariadne
    from .event.message import MessageEvent


class MessageChainDispatcher(BaseDispatcher):
    """从 MessageEvent 提取 MessageChain 的 Dispatcher"""

    @staticmethod
    async def catch(interface: DispatcherInterface["MessageEvent"]):
        if interface.annotation is MessageChain:
            return interface.event.messageChain


class ContextDispatcher(BaseDispatcher):
    """提取上下文的 Dispatcher"""

    @staticmethod
    async def catch(interface: DispatcherInterface):
        from asyncio import AbstractEventLoop

        from graia.broadcast import Broadcast

        from .adapter import Adapter
        from .app import Ariadne

        if not isinstance(interface.annotation, type):
            return
        if issubclass(interface.annotation, Ariadne):
            return ariadne_ctx.get()
        if issubclass(interface.annotation, Broadcast):
            return broadcast_ctx.get()
        if issubclass(interface.annotation, AbstractEventLoop):
            return event_loop_ctx.get()
        if issubclass(interface.annotation, Adapter):
            return adapter_ctx.get()
        if issubclass(interface.annotation, Dispatchable):
            return interface.event


class ContextLC(TypedDict):
    """local storage 表示"""

    __CONTEXT_MANAGER__: ContextManager


class MiddlewareDispatcher(BaseDispatcher):
    """分发 Ariadne 等基础参数的 Dispatcher"""

    def __init__(self, app: "Ariadne") -> None:
        self.app: "Ariadne" = app

    async def catch(self, interface: DispatcherInterface):
        return await ContextDispatcher.catch(interface)

    async def beforeExecution(self, interface: DispatcherInterface):
        """进入事件分发上下文"""
        lc: ContextLC = interface.execution_contexts[-1].local_storage
        lc["__CONTEXT_MANAGER__"] = enter_context(self.app, interface.event)
        lc["__CONTEXT_MANAGER__"].__enter__()

    async def afterExecution(self, interface: DispatcherInterface, *_):
        """退出事件分发上下文"""
        lc: ContextLC = interface.execution_contexts[-1].local_storage
        lc["__CONTEXT_MANAGER__"].__exit__(None, None, None)


class SourceDispatcher(BaseDispatcher):
    """提取 MessageEvent 消息链 Source 元素的 Dispatcher"""

    @staticmethod
    async def catch(interface: DispatcherInterface["MessageEvent"]):
        if interface.annotation is Source:
            return interface.event.messageChain.getFirst(Source)



"""Ariadne 消息相关的导入集合"""

# no error

from ..message.chain import MessageChain
from ..message.component import Component
from ..message.element import (
    App,
    At,
    AtAll,
    Dice,
    Element,
    Face,
    File,
    FlashImage,
    Forward,
    ForwardNode,
    Image,
    ImageType,
    MultimediaElement,
    MusicShare,
    NotSendableElement,
    Plain,
    Poke,
    PokeMethods,
    Quote,
    Source,
    Voice,
)
from ..message.formatter import Formatter
from ..message.parser.base import DetectPrefix, DetectSuffix
from ..message.parser.literature import (
    BoxParameter,
    Literature,
    ParamPattern,
    SwitchParameter,
)
from ..message.parser.twilight import (
    FORCE,
    NOSPACE,
    PRESERVE,
    ArgumentMatch,
    ElementMatch,
    FullMatch,
    Match,
    RegexMatch,
    Sparkle,
    Twilight,
    UnionMatch,
    WildcardMatch,
)



"""Ariadne 消息事件"""
from typing import Union

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface

from ..dispatcher import ContextDispatcher, MessageChainDispatcher, SourceDispatcher
from ..message.chain import MessageChain
from ..model import Client, Friend, Group, Member, Stranger
from . import MiraiEvent
from .mirai import FriendEvent, GroupEvent


class MessageEvent(MiraiEvent):
    """Ariadne 消息事件基类"""

    type: str = "MessageEvent"
    messageChain: MessageChain
    sender: Union[Friend, Member, Client, Stranger]

    class Dispatcher(BaseDispatcher):  # pylint: disable=missing-class-docstring
        mixin = [MessageChainDispatcher, ContextDispatcher, SourceDispatcher]

        @staticmethod
        async def catch(interface: DispatcherInterface):
            pass


class FriendMessage(MessageEvent, FriendEvent):
    """好友消息"""

    type: str = "FriendMessage"
    messageChain: MessageChain
    sender: Friend

    class Dispatcher(BaseDispatcher):  # pylint: disable=missing-class-docstring
        mixin = [MessageChainDispatcher, ContextDispatcher, SourceDispatcher]

        @staticmethod
        async def catch(interface: DispatcherInterface["FriendMessage"]):
            if interface.annotation is Friend:
                return interface.event.sender


class GroupMessage(MessageEvent, GroupEvent):
    """群组消息"""

    type: str = "GroupMessage"
    messageChain: MessageChain
    sender: Member

    class Dispatcher(BaseDispatcher):  # pylint: disable=missing-class-docstring
        mixin = [MessageChainDispatcher, ContextDispatcher, SourceDispatcher]

        @staticmethod
        async def catch(interface: DispatcherInterface["GroupMessage"]):
            if interface.annotation is Group:
                return interface.event.sender.group
            if interface.annotation is Member:
                return interface.event.sender


class TempMessage(MessageEvent):
    """临时消息"""

    type: str = "TempMessage"
    messageChain: MessageChain
    sender: Member

    class Dispatcher(BaseDispatcher):  # pylint: disable-next=missing-class-docstring

        mixin = [MessageChainDispatcher, ContextDispatcher, SourceDispatcher]

        @staticmethod
        async def catch(interface: DispatcherInterface["TempMessage"]):
            if interface.annotation is Group:
                return interface.event.sender.group
            if interface.annotation is Member:
                return interface.event.sender


class OtherClientMessage(MessageEvent):
    """其他客户端消息"""

    type: str = "OtherClientMessage"
    messageChain: MessageChain
    sender: Client

    class Dispatcher(BaseDispatcher):  # pylint: disable-next=missing-class-docstring

        mixin = [MessageChainDispatcher, ContextDispatcher, SourceDispatcher]

        @staticmethod
        async def catch(interface: DispatcherInterface["OtherClientMessage"]):
            if interface.annotation is Client:
                return interface.event.sender


class StrangerMessage(MessageEvent):
    """陌生人消息"""

    type: str = "StrangerMessage"
    messageChain: MessageChain
    sender: Stranger

    class Dispatcher(BaseDispatcher):  # pylint: disable-next=missing-class-docstring

        mixin = [MessageChainDispatcher, ContextDispatcher, SourceDispatcher]

        @staticmethod
        async def catch(interface: DispatcherInterface["StrangerMessage"]):
            if interface.annotation is Friend:
                return interface.event.sender



"""Alconna 的简单封装"""
from typing import TYPE_CHECKING

from arclet.alconna import Alconna, Arpamar
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface

from ...event.message import MessageEvent
from ..chain import MessageChain

if TYPE_CHECKING:
    ArpamarProperty = type("ArpamarProperty", (str, MessageChain), {})
else:
    ArpamarProperty = type("ArpamarProperty", tuple(), {})


class AlconnaDispatcher(BaseDispatcher):
    """
    Alconna的调度器形式
    """

    def __init__(self, *, alconna: Alconna):
        super().__init__()
        self.alconna = alconna

    async def beforeExecution(self, interface: "DispatcherInterface[MessageEvent]"):
        """预处理消息链并存入 local_storage"""
        local_storage = interface.execution_contexts[-1].local_storage
        chain: MessageChain = await interface.lookup_param("message_chain", MessageChain, None, [])
        result = self.alconna.analyse_message(chain)
        local_storage["arpamar"] = result

    async def catch(self, interface: DispatcherInterface[MessageEvent]) -> Arpamar:
        local_storage = interface.execution_contexts[-1].local_storage
        arpamar: Arpamar = local_storage["arpamar"]
        if issubclass(interface.annotation, ArpamarProperty):
            return arpamar.get(interface.name)
        if issubclass(interface.annotation, Arpamar):
            return arpamar
        if issubclass(interface.annotation, Alconna):
            return self.alconna



"""Ariadne 一站式导入的提供模块"""

# no error

from ..adapter import CombinedAdapter, DefaultAdapter, HttpAdapter, WebsocketAdapter
from ..app import Ariadne
from ..context import (
    adapter_ctx,
    ariadne_ctx,
    broadcast_ctx,
    event_ctx,
    event_loop_ctx,
    upload_method_ctx,
)
from ..dispatcher import ContextDispatcher, MessageChainDispatcher, SourceDispatcher
from ..exception import (
    AccountMuted,
    AccountNotFound,
    InvalidArgument,
    InvalidEventTypeDefinition,
    InvalidSession,
    InvalidVerifyKey,
    MessageTooLong,
    UnknownError,
    UnknownTarget,
    UnVerifiedSession,
)
from ..model import (
    BotMessage,
    CallMethod,
    ChatLogConfig,
    Client,
    DownloadInfo,
    FileInfo,
    Friend,
    Group,
    GroupConfig,
    Member,
    MemberInfo,
    MemberPerm,
    MiraiSession,
    Profile,
    Stranger,
    UploadMethod,
)
from ..util import cpu_bound, io_bound
from . import event, message



"""Ariadne 的类型标注"""
from typing import (
    AbstractSet,
    Any,
    Dict,
    Generic,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from typing_extensions import ParamSpec

P = ParamSpec("P")
R = TypeVar("R")
T = TypeVar("T")

Self = TypeVar("Self")

T_start = TypeVar("T_start")
T_stop = TypeVar("T_stop")
T_step = TypeVar("T_step")


class Slice(Generic[T_start, T_stop, T_step]):
    """对 slice 对象的泛型化包装, 但无法直接继承于 slice"""

    start: T_start
    stop: T_stop
    step: T_step


MessageIndex = Tuple[int, Optional[int]]

IntStr = Union[int, str]
AbstractSetIntStr = AbstractSet[IntStr]
DictIntStrAny = Dict[IntStr, Any]
DictStrAny = Dict[str, Any]
MappingIntStrAny = Mapping[IntStr, Any]
ReprArgs = Sequence[Tuple[Optional[str], Any]]



"""Ariadne 的异常定义"""


class InvalidEventTypeDefinition(Exception):
    """不合法的事件类型定义."""


class InvalidVerifyKey(Exception):
    """无效的 verifyKey 或其配置."""


class AccountNotFound(Exception):
    """未能使用所配置的账号激活 sessionKey, 请检查 mirai_session 配置."""


class InvalidSession(Exception):
    """无效的 sessionKey, 请重新获取."""


class UnVerifiedSession(Exception):
    """尚未验证/绑定的 session."""


class UnknownTarget(Exception):
    """对象位置未知, 不存在或不可及."""


class AccountMuted(Exception):
    """账号在对象所在聊天区域被封禁."""


class MessageTooLong(Exception):
    """消息过长, 尝试分段发送或报告问题."""


class InvalidArgument(Exception):
    """操作参数不合法, 请报告问题."""


class NotSupportedAction(Exception):
    """该版本不支持本接口."""


class DeprecatedImpl(Exception):
    """该接口已弃用."""


class MissingNecessaryArgument(Exception):
    """应在所提到的参数之中至少传入/使用一个"""


class ConflictItem(Exception):
    """项冲突/其中一项被重复定义"""


class UnknownError(Exception):
    """其他错误"""



"""Ariadne 基础的 parser, 包括 DetectPrefix 与 DetectSuffix"""
from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.exceptions import ExecutionStop
from graia.broadcast.interfaces.decorator import DecoratorInterface

from ..chain import MessageChain
from ..element import Quote, Source


class DetectPrefix(Decorator):
    """前缀检测器"""

    pre = True

    def __init__(self, prefix: str) -> None:
        """初始化前缀检测器.

        Args:
            prefix (str): 要匹配的前缀
        """
        self.prefix = prefix

    async def target(self, interface: DecoratorInterface):
        """检测前缀并 decorate 参数"""
        chain: MessageChain = await interface.dispatcher_interface.lookup_param(
            "message_chain", MessageChain, None, []
        )
        header = chain.include(Quote, Source)
        rest: MessageChain = chain.exclude(Quote, Source)
        if not rest.startswith(self.prefix):
            raise ExecutionStop
        result = rest.removeprefix(self.prefix)
        if interface.annotation is MessageChain:
            return header + result
        return None


class DetectSuffix(Decorator):
    """后缀检测器"""

    pre = True

    def __init__(self, suffix: str) -> None:
        """初始化后缀检测器.

        Args:
            suffix (str): 要匹配的后缀
        """
        self.suffix = suffix

    async def target(self, interface: DecoratorInterface):
        """检测后缀并 decorate 参数"""
        chain: MessageChain = await interface.dispatcher_interface.lookup_param(
            "message_chain", MessageChain, None, []
        )
        header = chain.include(Quote, Source)
        rest: MessageChain = chain.exclude(Quote, Source)
        if not rest.endswith(self.suffix):
            raise ExecutionStop
        result = rest.removesuffix(self.suffix)
        if interface.annotation is MessageChain:
            return header + result
        return None



"""Ariadne
一个优雅的 QQ Bot 框架.
"""
import graia.ariadne.event.lifecycle  # noqa: F401
import graia.ariadne.event.message  # noqa: F401
import graia.ariadne.event.mirai  # noqa: F401
import graia.ariadne.event.network  # noqa: F401

# init event

ARIADNE_ASCII_LOGO = "\n".join(
    (
        r"                _           _             ",
        r"     /\        (_)         | |            ",
        r"    /  \   _ __ _  __ _  __| |_ __   ___  ",
        r"   / /\ \ | '__| |/ _` |/ _` | '_ \ / _ \ ",
        r"  / ____ \| |  | | (_| | (_| | | | |  __/ ",
        r" /_/    \_\_|  |_|\__,_|\__,_|_| |_|\___| ",
        r"",
    )
)



"""Ariadne 事件相关的导入集合"""

# no error

from graia.ariadne.event import MiraiEvent
from graia.ariadne.event.lifecycle import (
    AdapterLaunched,
    AdapterShutdowned,
    ApplicationLaunched,
    ApplicationLifecycleEvent,
    ApplicationShutdowned,
)
from graia.ariadne.event.message import (
    FriendMessage,
    GroupMessage,
    OtherClientMessage,
    StrangerMessage,
    TempMessage,
)
from graia.ariadne.event.mirai import (
    BotEvent,
    BotGroupPermissionChangeEvent,
    BotInvitedJoinGroupRequestEvent,
    BotJoinGroupEvent,
    BotLeaveEventActive,
    BotLeaveEventKick,
    BotMuteEvent,
    BotOfflineEventActive,
    BotOfflineEventDropped,
    BotOfflineEventForce,
    BotOnlineEvent,
    BotReloginEvent,
    BotUnmuteEvent,
    CommandExecutedEvent,
    FriendEvent,
    FriendInputStatusChangedEvent,
    FriendNickChangedEvent,
    FriendRecallEvent,
    GroupAllowAnonymousChatEvent,
    GroupAllowConfessTalkEvent,
    GroupAllowMemberInviteEvent,
    GroupEntranceAnnouncementChangeEvent,
    GroupEvent,
    GroupMuteAllEvent,
    GroupNameChangeEvent,
    GroupRecallEvent,
    MemberCardChangeEvent,
    MemberHonorChangeEvent,
    MemberJoinEvent,
    MemberJoinRequestEvent,
    MemberLeaveEventKick,
    MemberLeaveEventQuit,
    MemberMuteEvent,
    MemberPerm,
    MemberPermissionChangeEvent,
    MemberSpecialTitleChangeEvent,
    MemberUnmuteEvent,
    NewFriendRequestEvent,
    NudgeEvent,
    OtherClientOfflineEvent,
    OtherClientOnlineEvent,
    RequestEvent,
)
from graia.ariadne.event.network import InvalidRequest, RemoteException



"""Twilight: 混合式消息链处理器"""
import abc
import enum
import re
import string
from argparse import Action
from copy import copy, deepcopy
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
    Union,
    final,
    overload,
)

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.exceptions import ExecutionStop
from graia.broadcast.interfaces.dispatcher import DispatcherInterface
from loguru import logger
from pydantic.utils import Representation

from ..chain import MessageChain
from ..element import Element
from .util import (
    CommandToken,
    ElementType,
    MessageChainType,
    TwilightParser,
    elem_mapping_ctx,
    split,
    tokenize_command,
    transform_regex,
)


class SpacePolicy(str, enum.Enum):
    """指示 RegexMatch 的尾随空格策略."""

    value: str
    NOSPACE = ""
    PRESERVE = "( )?"
    FORCE = "( )"


NOSPACE = SpacePolicy.NOSPACE
PRESERVE = SpacePolicy.PRESERVE
FORCE = SpacePolicy.FORCE


# ANCHOR: Match


class Match(abc.ABC, Representation):
    """匹配器的抽象基类."""

    pattern: str
    optional: bool
    help: str
    matched: Optional[bool]
    result: Optional["MessageChain"]

    def __init__(self, pattern, optional: bool = False, help: str = "", alt_help: str = "") -> None:
        self.pattern = pattern
        self.optional = optional
        self.help = help
        self.result = None
        self.matched = None
        self.alt_help = alt_help
        if self.__class__ == Match:
            raise ValueError("You can't instantiate Match class directly!")

    def __repr_args__(self):
        args = []
        if self.matched is not None:
            args.append(("matched", self.matched))
            args.append(("result", self.result))
        args.append(("pattern", self.pattern))
        return args


class RegexMatch(Match):
    """基础的正则表达式匹配."""

    regex_match: Optional[re.Match]
    space: SpacePolicy

    def __init__(
        self,
        pattern: str,
        *,
        optional: bool = False,
        flags: re.RegexFlag = re.RegexFlag(0),
        space: SpacePolicy = SpacePolicy.PRESERVE,
        help: str = "",
        alt_help: str = "",
        preserve_space: bool = ...,
    ) -> None:
        super().__init__(pattern=pattern, optional=optional, help=help, alt_help=alt_help)
        self.flags = flags
        self.regex_match = None
        self.space = space
        if preserve_space is not ...:
            logger.warning('"preserve_space argument" is deprecated and will be removed in 0.5.2!')
            logger.warning('use "space" instead!')
            self.space = SpacePolicy.PRESERVE if preserve_space else SpacePolicy.NOSPACE

    @final
    def gen_regex(self) -> str:
        """生成 RegexMatch 相应的正则表达式."""
        return (
            f"{transform_regex(self.flags, self.regex_src)}"
            f"{'?' if self.optional else ''}{self.space.value}"
        )

    @property
    def regex_src(self) -> str:
        """正则表达式的来源"""
        return self.pattern

    def get_help(self) -> str:
        """生成用于 `Sparkle.get_help()` 的描述性字符串."""
        return self.pattern.replace("( )?", " ") if not self.alt_help else self.alt_help

    def __repr_args__(self):
        return super().__repr_args__() + [("space", self.space.name)]


class ParamMatch(RegexMatch):
    """与 WildcardMatch 类似, 但需要至少一个字符. 且仅匹配用空格分开的一段. (不支持输入的转义处理, 但是可以处理不同引号包裹)"""

    num_tags: List[int]
    str_tags: List[str]

    def __init__(
        self,
        *tags: Union[str, int],
        optional: bool = False,
        flags: re.RegexFlag = re.RegexFlag(0),
        space: SpacePolicy = SpacePolicy.PRESERVE,
        help: str = "",
        alt_help: str = "",
        preserve_space: bool = ...,
    ) -> None:
        super().__init__(
            r"""(?:").+?(?:")|(?:').+?(?:')|\b[^ ]+?\b""",
            optional=optional,
            flags=flags,
            space=space,
            help=help,
            alt_help=alt_help,
            preserve_space=preserve_space,
        )
        self.num_tags: List[int] = list(filter(lambda x: isinstance(x, int), tags))
        self.str_tags: List[str] = list(filter(lambda x: isinstance(x, str), tags))

    def get_help(self) -> str:
        return "PARAM"

    def __repr_args__(self):
        return super().__repr_args__() + [("tags", self.num_tags + self.str_tags)]


class WildcardMatch(RegexMatch):
    """泛匹配."""

    greed: bool

    def __init__(
        self,
        *,
        greed: bool = True,
        optional: bool = False,
        flags: re.RegexFlag = re.RegexFlag(0),
        space: SpacePolicy = SpacePolicy.PRESERVE,
        help: str = "",
        alt_help: str = "",
        preserve_space: bool = ...,
    ) -> None:
        super().__init__(
            ".*",
            optional=optional,
            flags=flags,
            space=space,
            help=help,
            alt_help=alt_help,
            preserve_space=preserve_space,
        )
        self.greed = greed

    @property
    def regex_src(self) -> str:
        return f"{self.pattern}{'?' if not self.greed else ''}"


class FullMatch(RegexMatch):
    """全匹配."""

    @property
    def regex_src(self) -> str:
        return re.escape(self.pattern)

    def get_help(self) -> str:
        return self.pattern


class ElementMatch(RegexMatch):
    """元素类型匹配."""

    pattern: Type["Element"]
    result: "Element"

    def __init__(
        self,
        pattern: Type["Element"],
        *,
        optional: bool = False,
        flags: re.RegexFlag = re.RegexFlag(0),
        space: SpacePolicy = SpacePolicy.PRESERVE,
        help: str = "",
        alt_help: str = "",
        preserve_space: bool = ...,
    ) -> None:
        super().__init__(
            pattern,
            optional=optional,
            flags=flags,
            space=space,
            help=help,
            alt_help=alt_help,
            preserve_space=preserve_space,
        )

    @property
    def regex_src(self) -> str:
        return f"\x02\\d+_{self.pattern.__fields__['type'].default}\x03"

    def get_help(self) -> str:
        return self.pattern.__name__ if not self.alt_help else self.alt_help


class UnionMatch(RegexMatch):
    """多重匹配."""

    pattern: Tuple[str, ...]

    def __init__(
        self,
        *pattern: str,
        optional: bool = False,
        flags: re.RegexFlag = re.RegexFlag(0),
        space: SpacePolicy = SpacePolicy.PRESERVE,
        help: str = "",
        alt_help: str = "",
        preserve_space: bool = ...,
    ) -> None:
        super().__init__(
            pattern,
            optional=optional,
            flags=flags,
            space=space,
            help=help,
            alt_help=alt_help,
            preserve_space=preserve_space,
        )

    @property
    def regex_src(self) -> str:
        return f"{'|'.join(re.escape(i) for i in self.pattern)}"

    def get_help(self) -> str:
        return f"[{'|'.join(self.pattern)}]" if not self.alt_help else self.alt_help


class ArgumentMatch(Match):
    """参数匹配."""

    pattern: Sequence[str]
    name: str
    nargs: Union[str, int]
    action: Union[str, Type[Action]]
    dest: Optional[str]
    choices: Optional[Iterable]
    const: Any
    default: Any
    regex: Optional[re.Pattern]
    result: Union["MessageChain", Any]
    add_arg_data: Dict[str, Any]

    def __init__(
        self,
        *pattern: str,
        optional: bool = True,
        action: Union[str, Type[Action]] = ...,
        nargs: Union[int, str] = ...,
        const: Any = ...,
        default: Any = ...,
        type: Callable[[str], Any] = ...,
        choices: Optional[Iterable] = ...,
        help: Optional[str] = ...,
        dest: Optional[str] = ...,
        regex: Optional[str] = ...,
    ) -> None:
        if not pattern:
            raise ValueError("Expected at least 1 pattern!")
        super().__init__(pattern, optional, help)
        self.name = pattern[0].lstrip("-").replace("-", "_")
        self.nargs = nargs
        self.action = action
        self.const = const
        self.default = default
        self.choices = choices
        self.dest = dest
        self.regex = re.compile(regex) if regex is not ... else None
        data: Dict[str, Any] = {}
        if action is not ...:
            data["action"] = action
        if nargs is not ...:
            data["nargs"] = nargs
        if const is not ...:
            data["const"] = const
        if default is not ...:
            data["default"] = default
        if type is not ...:
            data["type"] = type
        if help is not ...:
            data["help"] = help
        if dest is not ...:
            data["dest"] = dest
        if choices is not ...:
            data["choices"] = choices
        if pattern[0].startswith("-"):
            data["required"] = not optional
        self.add_arg_data = data


T_Match = TypeVar("T_Match", bound=Match)

T_RMatch = TypeVar("T_RMatch", bound=RegexMatch)


# ANCHOR: Sparkle
class Sparkle(Representation):
    """Sparkle: Twilight 的匹配容器"""

    __dict__: Dict[str, Match]

    _description: str = ""
    _epilog: str = ""

    def __repr_args__(self):
        check = [(None, [item[0] for item in self._list_check_match])]
        return (
            check
            + [(k, v) for k, (v, _) in self._mapping_regex_match.items()]
            + list(self._mapping_arg_match.items())
        )

    @overload
    def __getitem__(self, item: Union[str, int]) -> Match:
        ...

    @overload
    def __getitem__(self, item: Type[T_Match]) -> List[T_Match]:
        ...

    @overload
    def __getitem__(self, item: Tuple[Type[T_RMatch], int]) -> T_RMatch:
        ...

    def __getitem__(self, item: Union[str, int, Type[T_Match], Tuple[Type[T_RMatch], int]]):
        return self.get_match(item)

    def __init_subclass__(cls, /, *, description: str = "", epilog: str = "", **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        cls._description = description
        cls._epilog = epilog

    def __getattribute__(self, _name: str):
        obj = super().__getattribute__(_name)
        if isinstance(obj, Match):
            return self.get_match(_name)
        return obj

    @overload
    def get_match(self, item: Union[str, int]) -> Match:
        ...

    @overload
    def get_match(self, item: Type[T_Match]) -> List[T_Match]:
        ...

    @overload
    def get_match(self, item: Tuple[T_RMatch, int]) -> T_RMatch:
        ...

    def get_param(self, tag: Union[int, str]) -> ParamMatch:
        """通过 tag 获取对应的 ParamMatch 实例

        Args:
            tag (Union[int, str]): ParamMatch 的 tag

        Returns:
            ParamMatch: 获取到的 ParamMatch

        """
        return self._param_match_ref[tag]

    def get_match(
        self, item: Union[str, int, Type[T_Match], Tuple[Type[T_RMatch], int]]
    ) -> Union[List[Match], Match]:
        """获取 Match 列表 / 实例.

        Args:
            item (Union[str, int, Type[T_Match], Tuple[Type[T_RMatch], int]]): 提供的信息, 请参见重载.

        Raises:
            KeyError: 找不到 Match.

        Returns:
            Union[List[Match], Match]: 获取的 Match 列表 / 实例.
        """
        if isinstance(item, int):
            return self._list_check_match[item][0]
        if isinstance(item, str):
            if item in self._mapping_arg_match:
                return self._mapping_arg_match[item]
            if item in self._mapping_regex_match:
                return self._mapping_regex_match[item][0]
        if isinstance(item, type):
            return self._match_ref[item]
        if isinstance(item, tuple):
            typ, ind = item
            return self._match_ref[typ][ind]
        raise KeyError(f"Unable to find match named {item}")

    def __deepcopy__(self, memo):
        copied = copy(self)

        copied._list_check_match = deepcopy(self._list_check_match, memo)
        copied._mapping_arg_match = deepcopy(self._mapping_arg_match, memo)
        copied._mapping_regex_match = deepcopy(self._mapping_regex_match, memo)
        copied._parser_ref = deepcopy(self._parser_ref, memo)
        copied._match_ref = deepcopy(self._match_ref, memo)
        copied._param_match_ref = deepcopy(self._param_match_ref, memo)

        return copied

    @overload
    def __init__(self, check: Dict[str, Match]):
        ...

    @overload
    def __init__(
        self,
        check: Iterable[RegexMatch] = (),
        match: Optional[Dict[str, Match]] = None,
    ):
        ...

    def __init__(
        self,
        check=(),
        match=None,
        description: str = "",
        epilog: str = "",
    ):
        self._description = description or self._description
        self._epilog = epilog or self._epilog

        check, match = (match, check) if isinstance(check, dict) else (check, match)

        check = check if check and check is not ... else ()
        match = match if match and match is not ... else ()

        match_map = {k: v for k, v in self.__class__.__dict__.items() if isinstance(v, Match)}
        match_map.update(match)

        # ----
        # ordinary matches
        # ----

        group_cnt: int = 0
        match_pattern_list: List[str] = []

        self._match_ref: DefaultDict[Type[T_Match], List[T_Match]] = DefaultDict(list)

        self._mapping_regex_match: Dict[str, Tuple[RegexMatch, int]] = {}
        self._mapping_arg_match: Dict[str, ArgumentMatch] = {}
        self._parser_ref: Dict[str, ArgumentMatch] = {}

        self._parser = TwilightParser(prog="", add_help=False)

        self._param_match_ref: Dict[Union[int, str], ParamMatch] = {}

        for v in check:
            self._match_ref[v.__class__].append(v)

            if isinstance(v, ParamMatch):  # Validate ParamMatch's tags
                for tag in v.num_tags + v.str_tags:
                    if tag in self._param_match_ref:
                        raise ValueError(f"Duplicated ParamMatch tag with {self._param_match_ref[tag]}")
                    self._param_match_ref[tag] = v

        for k, v in match_map.items():
            if k.startswith("_") or k[0] in string.digits:
                raise ValueError("Invalid Match object name!")

            self._match_ref[v.__class__].append(v)

            if isinstance(v, ArgumentMatch):  # add to self._parser
                self._mapping_arg_match[k] = v
                if v.action is ... or self._parser.accept_type(v.action):
                    if "type" not in v.add_arg_data or v.add_arg_data["type"] is MessageChain:
                        v.add_arg_data["type"] = MessageChainType(v.regex)
                    elif isinstance(v.add_arg_data["type"], type) and issubclass(
                        v.add_arg_data["type"], Element
                    ):
                        v.add_arg_data["type"] = ElementType(v.add_arg_data["type"])
                action = self._parser.add_argument(*v.pattern, **v.add_arg_data)
                v.dest = action.dest
                self._parser_ref[v.dest] = v
                continue

            if isinstance(v, RegexMatch):  # add to self._mapping_regex_match

                if isinstance(v, ParamMatch):  # Validate ParamMatch's tags
                    for tag in v.num_tags + v.str_tags:
                        if tag in self._param_match_ref:
                            raise ValueError(f"Duplicated ParamMatch tag with {self._param_match_ref[tag]}")
                        self._param_match_ref[tag] = v

                self._mapping_regex_match[k] = (v, group_cnt + 1)
                group_cnt += re.compile(v.gen_regex()).groups
                match_pattern_list.append(v.gen_regex())

                continue

            raise ValueError(f"{v} is neither RegexMatch nor ArgumentMatch!")

        if (
            not all(v.pattern[0].startswith("-") for v in self._mapping_arg_match.values())
            and self._mapping_regex_match
        ):  # inline validation for underscore
            raise ValueError("ArgumentMatch's pattern can't start with '-' in this case!")

        self._regex_pattern = "".join(match_pattern_list)
        self._regex = re.compile(self._regex_pattern)

        # ----
        # checking matches
        # ----

        self._list_check_match: List[Tuple[RegexMatch, int]] = []

        group_cnt = 0

        for check_match in check:
            if not isinstance(check_match, RegexMatch):
                raise ValueError(f"{check_match} can't be used as checking match!")
            self._list_check_match.append((check_match, group_cnt + 1))
            group_cnt += re.compile(check_match.gen_regex()).groups
        self._check_pattern: str = "".join(check_match.gen_regex() for check_match in check)
        self._check_regex = re.compile(self._check_pattern)

    @classmethod
    def from_command(  # ANCHOR: Sparkle: From command
        cls: "Type[Sparkle]", command: str, extra_arg_mapping: Optional[Dict[str, ArgumentMatch]] = None
    ) -> "Sparkle":
        """从 shell 式命令生成 Sparkle.

        Args:
            command (str): 命令, 使用 {0} 的形式创建参数占位符. 使用 [a|b] 创建选择匹配. 使用 反斜杠 转义.
            extra_arg_mapping (Dict[str, ArgumentMatch], optional): 可选的额外 str -> ArgumentMatch 映射.

        Returns:
            Sparkle: 生成的 Sparkle.
        """
        extra_arg_mapping = extra_arg_mapping or {}
        match: List[Union[FullMatch, ParamMatch, UnionMatch]] = []

        for t_type, token_list in tokenize_command(command):
            if t_type is CommandToken.TEXT:
                match.append(FullMatch(*token_list, space=SpacePolicy.FORCE))
            elif t_type is CommandToken.CHOICE:
                match.append(UnionMatch(*token_list, space=SpacePolicy.FORCE))
            elif t_type is CommandToken.PARAM:
                match.append(ParamMatch(*token_list, space=SpacePolicy.FORCE))
            else:
                raise ValueError(f"unexpected token type: {t_type}")

        if match:
            match[-1].space = SpacePolicy.NOSPACE

        return cls(match, extra_arg_mapping)

    # ANCHOR: Sparkle runtime populate

    def populate_check_match(self, string: str, elem_mapping: Dict[int, Element]) -> List[str]:
        """从传入的 string 与 elem_mapping 填充本实例的 check_match

        Args:
            string (str): 映射字符串
            elem_mapping (Dict[int, Element]): 元素映射

        Raises:
            ValueError: check_match 匹配失败

        Returns:
            List[str]: 剩下的字符串参数列表
        """
        if not self._check_pattern:
            return split(string)
        if regex_match := self._check_regex.match(string):
            for match, index in self._list_check_match:
                current = regex_match.group(index) or ""
                if isinstance(match, ElementMatch):
                    if current:
                        index = re.fullmatch("\x02(\\d+)_\\w+\x03", current).group(1)
                        result = elem_mapping[int(index)]
                    else:
                        result = None
                else:
                    result = MessageChain.fromMappingString(current, elem_mapping)

                match.result = result
                match.matched = bool(current)

                if match.__class__ is RegexMatch:
                    match.regex_match = re.fullmatch(match.pattern, current)
            return split(string[regex_match.end() :])
        raise ValueError(f"Not matching regex: {self._check_pattern}")

    def populate_arg_match(self, args: List[str]) -> List[str]:
        """从传入的 string 与填充本实例的 ArgumentMatch 对象

        Args:
            string (str): 映射字符串

        Returns:
            List[str]: 剩下的字符串参数列表
        """
        if not self._parser_ref:  # Optimization: skip if no ArgumentMatch
            return args
        namespace, rest = self._parser.parse_known_args(args)
        for arg_name, match in self._parser_ref.items():
            namespace_val = getattr(namespace, arg_name, ...)
            if namespace_val is not ...:
                match.result = namespace_val
                match.matched = bool(namespace_val)

        return rest

    def populate_regex_match(self, arg_list: List[str], elem_mapping: Dict[str, Element]) -> None:
        """从传入的 string 与 elem_mapping 填充本实例的 RegexMatch

        Args:
            string (str): 映射字符串
            elem_mapping (Dict[int, Element]): 元素映射

        Raises:
            ValueError: 匹配失败
        """
        if self._regex_pattern:
            if regex_match := self._regex.fullmatch(" ".join(arg_list)):
                for _, (match, index) in self._mapping_regex_match.items():
                    current = regex_match.group(index) or ""
                    if isinstance(match, ElementMatch):
                        if current:
                            index = re.fullmatch("\x02(\\d+)_\\w+\x03", current).group(1)
                            result = elem_mapping[int(index)]
                        else:
                            result = None
                    else:
                        result = MessageChain.fromMappingString(current, elem_mapping)

                    match.result = result
                    match.matched = bool(current)

                    if match.__class__ is RegexMatch:
                        match.regex_match = re.fullmatch(match.pattern, current)

            else:
                raise ValueError(f"Regex not matching: {self._regex_pattern}")

    def get_help(self, description: str = "", epilog: str = "", *, header: bool = True) -> str:
        """利用 Match 中的信息生成帮助字符串.

        Args:
            description (str, optional): 前导描述. Defaults to "".
            epilog (str, optional): 后置总结. Defaults to "".
            header (bool, optional): 生成使用方法 (命令总览). Defaults to True.

        Returns:
            str: 生成的帮助字符串, 被格式化与缩进过了
        """

        formatter = self._parser._get_formatter()

        description = description or self._description
        formatter.add_text(description)

        if header:
            header: List[str] = ["使用方法:"]

            for match, *_ in self._list_check_match:
                header.append(match.get_help())

            for match, *_ in self._mapping_regex_match.values():
                header.append(match.get_help())

            formatter.add_usage(None, self._parser._actions, [], prefix=" ".join(header) + " ")

        positional, optional, *_ = self._parser._action_groups
        formatter.start_section("位置匹配")
        for name, (match, _) in self._mapping_regex_match.items():
            formatter.add_text(f"{name} -> 匹配 {match.get_help()}{' : ' + match.help if match.help else ''}")
        formatter.add_arguments(positional._group_actions)
        formatter.end_section()

        formatter.start_section("参数匹配")
        formatter.add_arguments(optional._group_actions)
        formatter.end_section()

        epilog = epilog or self._epilog
        formatter.add_text(epilog)

        # determine help from format above
        return formatter.format_help()


T_Sparkle = TypeVar("T_Sparkle", bound=Sparkle)


class _TwilightLocalStorage(TypedDict):
    result: Optional[Sparkle]


# ANCHOR: Twilight
class Twilight(BaseDispatcher, Generic[T_Sparkle]):
    """
    暮光.
    """

    @overload
    def __init__(self, root: Dict[str, Match], *, map_params: Optional[Dict[str, bool]] = None):
        """本魔法方法用于初始化本实例.

        Args:
            check (Dict[str, Match]): 匹配的映射.
            map_params (Dict[str, bool], optional): 向 MessageChain.asMappingString 传入的参数.
        """

    @overload
    def __init__(
        self, root: Union[Type[T_Sparkle], T_Sparkle], *, map_params: Optional[Dict[str, bool]] = None
    ):
        """本魔法方法用于初始化本实例.

        Args:
            root (Union[Type[Twilight], Twilight], optional): 根 Sparkle 实例, 用于生成新的 Sparkle.
            map_params (Dict[str, bool], optional): 向 MessageChain.asMappingString 传入的参数.
        """

    @overload
    def __init__(
        self,
        root: Iterable[RegexMatch],
        match: Dict[str, Match] = ...,
        *,
        map_params: Optional[Dict[str, bool]] = None,
    ):
        """本魔法方法用于初始化本实例.

        Args:
            check (Iterable[RegexMatch]): 用于检查的 Match 对象.
            match (Dict[str, Match]): 额外匹配的映射.
            map_params (Dict[str, bool], optional): 向 MessageChain.asMappingString 传入的参数.
        """

    def __init__(self, root=..., match=..., *, map_params: Optional[Dict[str, bool]] = None):
        """Actual implementation of __init__"""
        if isinstance(root, Sparkle):
            self.root = root
        elif isinstance(root, type) and issubclass(root, Sparkle):
            self.root = root()
        else:
            self.root = Sparkle(check=root, match=match)

        self._map_params = map_params or {}

    @classmethod
    def from_command(
        cls,
        command: str,
        extra_arg_mapping: Optional[Dict[str, ArgumentMatch]] = None,
        *,
        map_params: Optional[Dict[str, bool]] = None,
    ) -> "Twilight":
        """从 shell 式命令生成 Twilight.

        Args:
            command (str): 命令, 使用 {0} 的形式创建参数占位符.
            extra_arg_mapping (Dict[str, ArgumentMatch], optional): 可选的额外 str -> ArgumentMatch 映射.
            map_params (Dict[str, bool], optional): 向 MessageChain.asMappingString 传入的参数.

        Returns:
            Twilight: 生成的 Twilight.
        """
        return cls(Sparkle.from_command(command, extra_arg_mapping), map_params=map_params)

    def generate(self, chain: MessageChain) -> T_Sparkle:
        """从消息链生成 Sparkle 实例.

        Args:
            chain (MessageChain): 传入的消息链.

        Returns:
            T_Sparkle: 生成的 Sparkle 对象.
        """
        sparkle = deepcopy(self.root)
        mapping_str, elem_mapping = chain.asMappingString(**self._map_params)
        token = elem_mapping_ctx.set(chain)
        try:
            str_list = sparkle.populate_check_match(mapping_str, elem_mapping)
            arg_list = sparkle.populate_arg_match(str_list)
            sparkle.populate_regex_match(arg_list, elem_mapping)
            return sparkle
        finally:
            elem_mapping_ctx.reset(token)

    async def beforeExecution(self, interface: DispatcherInterface):
        """检验 MessageChain 并将 Sparkle 存入本地存储

        Args:
            interface (DispatcherInterface): DispatcherInterface, 应该能从中提取 MessageChain

        Raises:
            ExecutionStop: 匹配以任意方式失败
        """
        local_storage: _TwilightLocalStorage = interface.execution_contexts[-1].local_storage
        chain: MessageChain = await interface.lookup_param("message_chain", MessageChain, None, [])
        try:
            local_storage["result"] = self.generate(chain)
        except Exception as e:
            raise ExecutionStop from e

    async def catch(self, interface: DispatcherInterface) -> Optional[Union["Twilight", T_Sparkle, Match]]:
        local_storage: _TwilightLocalStorage = interface.execution_contexts[-1].local_storage
        sparkle = local_storage["result"]
        if issubclass(interface.annotation, Sparkle):
            return sparkle
        if issubclass(interface.annotation, Twilight):
            return self
        if issubclass(interface.annotation, Match):
            return sparkle.get_match(interface.name)



"""Ariadne 各种 model 存放的位置"""
import json
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Any, Optional, Union

from loguru import logger
from pydantic import BaseModel, Field, validator
from pydantic.main import BaseConfig, Extra
from pydantic.networks import AnyHttpUrl
from typing_extensions import Literal
from yarl import URL

if TYPE_CHECKING:
    from .app import Ariadne
    from .message.chain import MessageChain
    from .typing import AbstractSetIntStr, DictStrAny, MappingIntStrAny


def datetime_encoder(v: datetime) -> float:
    """编码 datetime 对象

    Args:
        v (datetime): datetime 对象

    Returns:
        float: 编码后的 datetime (时间戳)
    """
    return v.timestamp()


class DatetimeEncoder(json.JSONEncoder):
    """可以编码 datetime 的 JSONEncoder"""

    def default(self, o: Any) -> Any:
        if isinstance(o, datetime):
            return int(o.timestamp())
        return super().default(o)


class AriadneBaseModel(BaseModel):
    """
    Ariadne 一切数据模型的基类.
    """

    def dict(
        self,
        *,
        include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        by_alias: bool = False,
        skip_defaults: bool = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> "DictStrAny":
        _, _ = by_alias, exclude_none
        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=True,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=True,
        )

    class Config(BaseConfig):
        """Ariadne BaseModel 设置"""

        extra = Extra.allow
        json_encoders = {
            datetime: datetime_encoder,
        }


@dataclass
class ChatLogConfig:
    """配置日志如何记录 QQ 消息与事件."""

    enabled: bool = True
    log_level: str = "INFO"
    group_message_log_format: str = (
        "{bot_id}: [{group_name}({group_id})] {member_name}({member_id}) -> {message_string}"
    )
    friend_message_log_format: str = "{bot_id}: [{friend_name}({friend_id})] -> {message_string}"
    temp_message_log_format: str = (
        "{bot_id}: [{group_name}({group_id}).{member_name}({member_id})] -> {message_string}"
    )
    other_client_message_log_format: str = "{bot_id}: [{platform_name}({platform_id})] -> {message_string}"
    stranger_message_log_format: str = "{bot_id}: [{stranger_name}({stranger_id})] -> {message_string}"

    def initialize(self, app: "Ariadne"):
        """利用 Ariadne 对象注册事件日志处理器"""
        from .event.message import (
            FriendMessage,
            GroupMessage,
            OtherClientMessage,
            StrangerMessage,
            TempMessage,
        )

        @app.broadcast.receiver(GroupMessage)
        def log_group_message(event: GroupMessage):
            logger.log(
                self.log_level,
                self.group_message_log_format.format(
                    group_id=event.sender.group.id,
                    group_name=event.sender.group.name,
                    member_id=event.sender.id,
                    member_name=event.sender.name,
                    member_permission=event.sender.permission.name,
                    bot_id=app.mirai_session.account,
                    message_string=event.messageChain.asDisplay().__repr__(),
                ),
            )

        @app.broadcast.receiver(FriendMessage)
        def log_friend_message(event: FriendMessage):
            logger.log(
                self.log_level,
                self.friend_message_log_format.format(
                    bot_id=app.mirai_session.account,
                    friend_name=event.sender.nickname,
                    friend_id=event.sender.id,
                    message_string=event.messageChain.asDisplay().__repr__(),
                ),
            )

        @app.broadcast.receiver(TempMessage)
        def log_temp_message(event: TempMessage):
            logger.log(
                self.log_level,
                self.temp_message_log_format.format(
                    group_id=event.sender.group.id,
                    group_name=event.sender.group.name,
                    member_id=event.sender.id,
                    member_name=event.sender.name,
                    member_permission=event.sender.permission.name,
                    bot_id=app.mirai_session.account,
                    message_string=event.messageChain.asDisplay().__repr__(),
                ),
            )

        @app.broadcast.receiver(StrangerMessage)
        def log_stranger_message(event: StrangerMessage):
            logger.log(
                self.log_level,
                self.stranger_message_log_format.format(
                    bot_id=app.mirai_session.account,
                    stranger_name=event.sender.nickname,
                    stranger_id=event.sender.id,
                    message_string=event.messageChain.asDisplay().__repr__(),
                ),
            )

        @app.broadcast.receiver(OtherClientMessage)
        def log_other_client_message(event: OtherClientMessage):
            logger.log(
                self.log_level,
                self.other_client_message_log_format.format(
                    bot_id=app.mirai_session.account,
                    platform_name=event.sender.platform,
                    platform_id=event.sender.id,
                    message_string=event.messageChain.asDisplay().__repr__(),
                ),
            )


class MiraiSession(AriadneBaseModel):
    """
    用于描述与上游接口会话, 并存储会话状态的实体类.

    Attributes:
        host (AnyHttpUrl): `mirai-api-http` 服务所在的根接口地址
        account (int): 应用所使用账号的整数 ID, 虽然启用 `singleMode` 时不需要, 但仍然建议填写.
        verify_key (str): 在 `mirai-api-http` 配置流程中定义, 需为相同的值以通过安全验证, 需在 mirai-api-http 配置里启用 `enableVerify`.
        session_key (str, optional): 会话标识, 即会话中用于进行操作的唯一认证凭证.
    """

    host: AnyHttpUrl
    single_mode: bool = False
    account: Optional[int] = None
    verify_key: Optional[str] = None
    session_key: Optional[str] = None
    version: Optional[str] = None

    def __init__(
        self,
        host: Union[AnyHttpUrl, str],
        account: Optional[Union[int, str]] = None,
        verify_key: Optional[str] = None,
        *,
        single_mode: bool = False,
    ) -> None:
        super().__init__(host=host, account=account, verify_key=verify_key, single_mode=single_mode)

    def url_gen(self, route: str) -> str:
        """生成 route 对应的 API URI

        Args:
            route (str): route 地址

        Returns:
            str: 对应的 API URI
        """
        return str(URL(self.host) / route)


class Friend(AriadneBaseModel):
    """描述 Tencent QQ 中的好友."""

    id: int
    nickname: str
    remark: str


class Stranger(AriadneBaseModel):
    """描述 Tencent QQ 中的陌生人."""

    id: int
    nickname: str
    remark: str


class MemberPerm(Enum):
    """描述群成员在群组中所具备的权限"""

    Member = "MEMBER"  # 普通成员
    Administrator = "ADMINISTRATOR"  # 管理员
    Owner = "OWNER"  # 群主


class Group(AriadneBaseModel):
    """描述 Tencent QQ 中的群组."""

    id: int
    name: str
    accountPerm: MemberPerm = Field(..., alias="permission")


class Member(AriadneBaseModel):
    """描述用户在群组中所具备的有关状态, 包括所在群组, 群中昵称, 所具备的权限, 唯一ID."""

    id: int
    name: str = Field(..., alias="memberName")
    permission: MemberPerm
    specialTitle: Optional[str] = None
    joinTimestamp: Optional[int] = None
    lastSpeakTimestamp: Optional[int] = None
    mutetimeRemaining: Optional[int] = None
    group: Group


class GroupConfig(AriadneBaseModel):
    """描述群组各项功能的设置."""

    name: str = ""
    announcement: str = ""
    confessTalk: bool = False
    allowMemberInvite: bool = False
    autoApprove: bool = False
    anonymousChat: bool = False


class MemberInfo(AriadneBaseModel):
    """描述群组成员的可修改状态, 修改需要管理员/群主权限."""

    name: str = ""
    specialTitle: str = ""


class DownloadInfo(AriadneBaseModel):
    """描述一个文件的下载信息."""

    sha: str = ""
    md5: str = ""
    download_times: int = Field(..., alias="downloadTimes")
    uploader_id: int = Field(..., alias="uploaderId")
    upload_time: datetime = Field(..., alias="uploadTime")
    last_modify_time: datetime = Field(..., alias="lastModifyTime")
    url: Optional[str] = None


class FileInfo(AriadneBaseModel):
    """群组文件详细信息"""

    name: str = ""
    path: str = ""
    id: Optional[str] = ""
    parent: Optional["FileInfo"] = None
    contact: Optional[Union[Group, Friend]] = None
    is_file: bool = Field(..., alias="isFile")
    is_directory: bool = Field(..., alias="isDirectory")
    download_info: Optional[DownloadInfo] = Field(None, alias="downloadInfo")

    @validator("contact", pre=True, allow_reuse=True)
    def _(cls, val: Optional[dict]):
        if not val:
            return None
        if "remark" in val:  # Friend
            return Friend.parse_obj(val)
        return Group.parse_obj(val)  # Group


FileInfo.update_forward_refs(FileInfo=FileInfo)


class UploadMethod(Enum):
    """用于向 `upload` 系列方法描述上传类型"""

    Friend = "friend"
    Group = "group"
    Temp = "temp"


class CallMethod(Enum):
    """
    用于向 `Adapter.call_api` 指示操作类型.
    """

    GET = "GET"
    POST = "POST"
    RESTGET = "get"
    RESTPOST = "update"
    MULTIPART = "multipart"


class Client(AriadneBaseModel):
    """
    指示其他客户端
    """

    id: int
    platform: str


class Profile(AriadneBaseModel):
    """
    指示某个用户的个人资料
    """

    nickname: str
    email: Optional[str]
    age: Optional[int]
    level: int
    sign: str
    sex: Literal["UNKNOWN", "MALE", "FEMALE"]


class BotMessage(AriadneBaseModel):
    """
    指示 Bot 发出的消息.
    """

    messageId: int

    origin: Optional["MessageChain"]


class AriadneStatus(Enum):
    """指示 Ariadne 状态的枚举类"""

    STOP = "stop"
    LAUNCH = "launch"
    RUNNING = "running"
    SHUTDOWN = "shutdown"
    CLEANUP = "cleanup"



"""消息链处理器的 "模式", 将在 0.5.2 移除"""
from loguru import logger

from .literature import BoxParameter, ParamPattern, SwitchParameter  # noqa: F401
from .twilight import (  # noqa: F401
    ArgumentMatch,
    ElementMatch,
    FullMatch,
    Match,
    RegexMatch,
    UnionMatch,
    WildcardMatch,
)

logger.warning("This module is deprecated and will be removed in 0.5.2!")



"""本模块提供 Ariadne 内部使用的小工具, 以及方便的 `async_exec` 模块.
"""

# Utility Layout
import asyncio
import functools
import sys
import traceback
import warnings
from asyncio.events import AbstractEventLoop
from typing import (
    AsyncIterator,
    Callable,
    Coroutine,
    Dict,
    Generator,
    List,
    Type,
    TypeVar,
    Union,
)

from graia.broadcast import Broadcast
from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.entities.event import Dispatchable
from graia.broadcast.entities.listener import Listener
from graia.broadcast.entities.namespace import Namespace
from graia.broadcast.typing import T_Dispatcher
from loguru import logger

from ..exception import (
    AccountMuted,
    AccountNotFound,
    InvalidArgument,
    InvalidSession,
    InvalidVerifyKey,
    MessageTooLong,
    UnknownError,
    UnknownTarget,
    UnVerifiedSession,
)
from ..typing import P, R, T

# Import layout
from . import async_exec  # noqa: F401
from .async_exec import (  # noqa: F401
    IS_MAIN_PROCESS,
    ParallelExecutor,
    cpu_bound,
    io_bound,
)

code_exceptions_mapping: Dict[int, Type[Exception]] = {
    1: InvalidVerifyKey,
    2: AccountNotFound,
    3: InvalidSession,
    4: UnVerifiedSession,
    5: UnknownTarget,
    6: FileNotFoundError,
    10: PermissionError,
    20: AccountMuted,
    30: MessageTooLong,
    400: InvalidArgument,
}


def validate_response(code: Union[dict, int]):
    """验证远程服务器的返回值

    Args:
        code (Union[dict, int]): 返回的对象

    Raises:
        Exception: 请参照 code_exceptions_mapping
    """
    origin = code
    if isinstance(code, dict):
        code = code.get("code")
    if not isinstance(code, int) or code == 200 or code == 0:
        return
    exc_cls = code_exceptions_mapping.get(code)
    if exc_cls:
        raise exc_cls(exc_cls.__doc__)
    raise UnknownError(f"{origin}")


def loguru_excepthook(cls, val, tb, *_, **__):
    """loguru 异常回调

    Args:
        cls (Type[Exception]): 异常类
        val (Exception): 异常的实际值
        tb (TracebackType): 回溯消息
    """
    logger.opt(exception=(cls, val, tb)).error("Exception:")


def loguru_async_handler(_, ctx: dict):
    """loguru 异步异常回调

    Args:
        _ (AbstractEventLoop): 异常发生的事件循环
        ctx (dict): 异常上下文
    """
    if "exception" in ctx:
        logger.opt(exception=ctx["exception"]).error("Exception:")
    else:
        logger.error(f"Exception: {ctx}")


def inject_loguru_traceback(loop: AbstractEventLoop = None):
    """使用 loguru 模块替换默认的 traceback.print_exception 与 sys.excepthook"""
    traceback.print_exception = loguru_excepthook
    sys.excepthook = loguru_excepthook
    if loop:
        loop.set_exception_handler(loguru_async_handler)


def inject_bypass_listener(broadcast: Broadcast):
    """注入 BypassListener 以享受子事件解析.

    Args:
        broadcast (Broadcast): 外部事件系统, 提供了 event_class_generator 方法以生成子事件.
    """

    class BypassListener(Listener):
        """透传监听器的实现"""

        def __init__(
            self,
            callable: Callable,
            namespace: Namespace,
            listening_events: List[Type[Dispatchable]],
            inline_dispatchers: List[T_Dispatcher] = None,
            decorators: List[Decorator] = None,
            priority: int = 16,
        ) -> None:
            events = []
            for event in listening_events:
                events.append(event)
                events.extend(broadcast.event_class_generator(event))
            super().__init__(
                callable,
                namespace,
                events,
                inline_dispatchers=inline_dispatchers,
                decorators=decorators,
                priority=priority,
            )

    import graia.broadcast.entities.listener

    # pylint: disable=no-member
    graia.broadcast.entities.listener.Listener = BypassListener  # type: ignore
    graia.broadcast.Listener = BypassListener  # type: ignore
    try:  # Override saya listener
        import graia.saya.builtins.broadcast.schema

        graia.saya.builtins.broadcast.schema.Listener = BypassListener  # type: ignore
    except ImportError:  # Saya not installed, pass.
        pass


def app_ctx_manager(func: Callable[P, R]) -> Callable[P, R]:
    """包装声明需要在 Ariadne Context 中执行的函数

    Args:
        func (Callable[P, R]): 被包装的函数

    Returns:
        Callable[P, R]: 包装后的函数
    """

    @functools.wraps(func)
    async def wrapper(self, *args: P.args, **kwargs: P.kwargs):
        from ..context import enter_context

        with enter_context(app=self):
            return await func(self, *args, **kwargs)

    return wrapper


def gen_subclass(cls: Type[T]) -> Generator[Type[T], None, None]:
    """生成某个类的所有子类 (包括其自身)

    Args:
        cls (Type[T]): 类

    Yields:
        Type[T]: 子类
    """
    yield cls
    for sub_cls in cls.__subclasses__():
        yield from gen_subclass(sub_cls)


def wrap_bracket(string: str) -> str:
    """替换 "[" 与 "]" 为其 Unicode 形式用于 JSON 解析"""
    return string.replace("[", "\\u005b").replace("]", "\\u005d")


T_Callable = TypeVar("T_Callable", bound=Callable)


async def await_predicate(predicate: Callable[[], bool], interval: float = 0.01) -> None:
    """异步阻塞至满足 predicate 为 True

    Args:
        predicate (Callable[[], bool]): 回调函数
        interval (float, optional): 每次等待时长 (s). Defaults to 0.01.
    """
    while not predicate():
        await asyncio.sleep(interval)


async def yield_with_timeout(
    getter_coro: Callable[[], Coroutine[None, None, T]],
    predicate: Callable[[], bool],
    await_length: float = 0.2,
) -> AsyncIterator[T]:
    """在满足 predicate 时返回 getter_coro() 的值

    Args:
        getter_coro (Callable[[], Coroutine[None, None, T]]): 要循环返回协程函数.
        predicate (Callable[[], bool]): 条件回调函数.
        await_length (float, optional): 等待目前协程的时长. Defaults to 0.2.

    Yields:
        T: getter_coro 的返回值
    """
    last_tsk = None
    while predicate():
        last_tsk = last_tsk or {asyncio.create_task(getter_coro())}
        done, last_tsk = await asyncio.wait(last_tsk, timeout=await_length)
        if not done:
            continue
        for t in done:
            res = await t
            yield res
    if last_tsk:
        for tsk in last_tsk:
            tsk.cancel()


def deprecated(remove_ver: str) -> Callable[[T_Callable], T_Callable]:
    """标注一个方法 / 函数已被弃用

    Args:
        remove_ver (str): 将被移除的版本.

    Returns:
        Callable[[T_Callable], T_Callable]: 包装器.
    """

    def out_wrapper(func: T_Callable) -> T_Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(DeprecationWarning(f"{func.__qualname__} will be removed in {remove_ver}!"))
            logger.warning(f"Deprecated function: {func.__qualname__}")
            logger.warning(f"{func.__qualname__} will be removed in {remove_ver}!")
            return func(*args, **kwargs)

        return wrapper

    return out_wrapper


class Dummy:
    """Dummy 类, 对所有调用返回 None. (可以预设某些值)"""

    def __init__(self, **kwds):
        for k, v in kwds:
            self.__setattr__(k, v)

    def __getattr__(self, *_, **__):
        return self

    def __call__(self, *_, **__):
        return self
